﻿using System;
using System.Net.Sockets;
using System.Text;
using EonaCat.LogSystem;
using EonaCat.Quic;
using EonaCat.Quic.Connections;
using EonaCat.Quic.Events;
using EonaCat.Quic.Helpers;
using EonaCat.Quic.Streams;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    public static class NetworkHelper
    {
        internal static Logging Logger = new Logging();
        private static QuicServer _quicServer;

        /// <summary>
        /// OnQuicClientConnected event
        /// </summary>
        public static event EventHandler<QuicConnectionEventArgs> OnQuicClientConnected;

        /// <summary>
        /// OnQuicStreamOpened event
        /// </summary>
        public static event EventHandler<QuicStreamEventArgs> OnQuicStreamOpened;

        /// <summary>
        /// OnQuicStreamDataReceived event
        /// </summary>
        public static event EventHandler<QuicStreamEventArgs> OnQuicStreamDataReceived;

        /// <summary>
        /// Start a Quic server
        /// </summary>
        /// <returns><c>true</c>, start successfully, <c>false</c> did not start successfully.</returns>
        /// <param name="ip">The ip bound to the server</param>
        /// <param name="port">The listening port. (default: 11000)</param>
        public static bool QuicStartServer(string ip, int port = 11000)
        {
            _quicServer = new QuicServer(ip, port);
            _quicServer.OnClientConnected += ClientConnected;
            _quicServer.Start();
            Logger.Info($"The Quic server has been successfully started on ip '{ip}' and port: {port}");
            return true;
        }

        /// <summary>
        /// Fired when Client is connected
        /// </summary>
        /// <param name="connection">The new connection</param>
        private static void ClientConnected(QuicConnection connection)
        {
            OnQuicClientConnected?.Invoke(null, new QuicConnectionEventArgs { Connection = connection });
            connection.OnStreamOpened += StreamOpened;
        }

        private static void StreamOpened(QuicStream stream)
        {
            OnQuicStreamOpened?.Invoke(null, new QuicStreamEventArgs { Stream = stream });
            stream.OnStreamDataReceived += StreamDataReceived;
        }

        private static void StreamDataReceived(QuicStream stream, byte[] data)
        {
            OnQuicStreamDataReceived?.Invoke(null, new QuicStreamEventArgs { Stream = stream, Data = data });
        }

        /// <summary>
        /// Start a Quic client
        /// </summary>
        /// <param name="ip"></param>
        /// <param name="port"></param>
        /// <param name="streamType"></param>
        /// <returns></returns>
        public static QuicStream QuicStartClient(string ip, int port = 11000, StreamType streamType = StreamType.ClientBidirectional)
        {
            QuicClient client = new QuicClient();

            // Connect to peer (Server)
            QuicConnection connection = client.Connect(ip, port);

            // Create a data stream
            return connection.CreateStream(streamType);
        }

        /// <summary>
        /// Stop the Quic server
        /// </summary>
        /// <returns></returns>
        public static bool QuicStopServer()
        {
            if (_quicServer != null)
            {
                _quicServer.Close();
                Logger.Info($"The Quic server has been successfully stopped");
            }
            return true;
        }

        //TODO: add udp and tcp example methods

        /// <summary>
        /// Character bit encoding type web
        /// </summary>
        public static Encoding GlobalEncoding = Encoding.UTF8;

        private static void Main(string[] args)
        {
            var client = new UdpClient();

            for (int i = 0; i < 5000; i++)
            {
                // TCP TEST
            }

            for (int i = 0; i < 5000; i++)
            {
                // UDP TEST
            }
            Console.ReadLine();
        }
    }

    /// <summary>
    /// IP type enumeration
    /// </summary>
    public enum IPType : byte
    {
        IPv4,
        IPv6
    }
}