﻿using System;
using System.ComponentModel;
using System.Net;
using System.Net.Sockets;
using System.Threading.Tasks;

namespace EonaCat.Network
{

    public class SocketTcpClient
    {
        private const int BUFFER_SIZE = 4096;

        /// <summary>
        /// OnConnect event
        /// </summary>
        public event Action<RemoteInfo> OnConnect;

        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<RemoteInfo> OnReceive;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<RemoteInfo> OnDisconnect;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        private Socket socket;

        /// <summary>
        /// Create TCP client
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        /// <returns></returns>
        /// <exception cref="Exception"></exception>
        public Task ConnectAsync(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            return CreateSocketTcpClientAsync(ip, port);
        }

        private async Task CreateSocketTcpClientAsync(IPAddress ipAddress, int port)
        {
            IsIp6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            socket = new Socket(IsIp6 ? AddressFamily.InterNetworkV6 : AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            try
            {
                await socket.ConnectAsync(ipAddress, port).ConfigureAwait(false);
                OnConnect?.Invoke(new RemoteInfo { IsTcp = true, IsIpv6 = IsIp6 });
                _ = StartReceivingAsync();
            }
            catch (SocketException ex)
            {
                OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                Disconnect();
            }
        }

        public bool IsIp6 { get; set; }

        private async Task StartReceivingAsync()
        {
            byte[] buffer = new byte[BUFFER_SIZE]; // Increased buffer size for better performance
            while (socket.Connected)
            {
                try
                {
                    int received = await socket.ReceiveAsync(new ArraySegment<byte>(buffer), SocketFlags.None).ConfigureAwait(false);
                    if (received > 0)
                    {
                        byte[] data = new byte[received];
                        Buffer.BlockCopy(buffer, 0, data, 0, received);
                        OnReceive?.Invoke(new RemoteInfo
                        {
                            IsTcp = true,
                            Data = data,
                            EndPoint = socket.RemoteEndPoint,
                            IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6
                        });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }
            OnDisconnect?.Invoke(new RemoteInfo { IsTcp = true, EndPoint = socket.RemoteEndPoint, IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6 });
        }

        /// <summary>
        /// Send data
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public Task SendAsync(byte[] data)
        {
            return socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None);
        }

        /// <summary>
        /// Disconnect
        /// </summary>
        public void Disconnect()
        {
            socket.Close();
        }
    }
}
