﻿using System.Net.Sockets;
using System.Net;
using System;
using System.Threading.Tasks;
using System.Threading;

namespace EonaCat.Network
{
    public class SocketTcpServer
    {
        private const int BUFFER_SIZE = 4096;

        /// <summary>
        /// OnConnect event
        /// </summary>
        public event Action<RemoteInfo> OnConnect;

        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<RemoteInfo> OnReceive;

        /// <summary>
        /// OnSend event
        /// </summary>
        public event Action<RemoteInfo> OnSend;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<RemoteInfo> OnDisconnect;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        private readonly TcpListener _listener;
        private CancellationTokenSource _cancellationTokenSource;
        private Task _acceptTask;

        /// <summary>
        /// Create TCP server
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        public SocketTcpServer(IPAddress ipAddress, int port)
        {
            _listener = new TcpListener(ipAddress, port);
        }

        /// <summary>
        /// Create TCP server
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        public SocketTcpServer(string ipAddress, int port)
        {
            IPAddress address = IPAddress.Parse(ipAddress);
            _listener = new TcpListener(address, port);
        }

        /// <summary>
        /// Start TCP server
        /// </summary>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        public Task StartAsync(CancellationToken cancellationToken = default)
        {
            _listener.Start();
            _cancellationTokenSource = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            _acceptTask = AcceptConnectionsAsync(_cancellationTokenSource.Token);
            return _acceptTask;
        }

        private async Task AcceptConnectionsAsync(CancellationToken cancellationToken)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var socket = await _listener.AcceptSocketAsync().ConfigureAwait(false);
                    _ = HandleConnectionAsync(socket, cancellationToken);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                }
            }
        }

        private async Task HandleConnectionAsync(Socket socket, CancellationToken cancellationToken)
        {
            OnConnect?.Invoke(new RemoteInfo
            {
                Socket = socket,
                IsTcp = true,
                IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6,
                EndPoint = socket.RemoteEndPoint
            });

            byte[] buffer = new byte[BUFFER_SIZE];
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    int received = await socket.ReceiveAsync(new Memory<byte>(buffer), SocketFlags.None, cancellationToken).ConfigureAwait(false);
                    if (received > 0)
                    {
                        byte[] data = new byte[received];
                        Buffer.BlockCopy(buffer, 0, data, 0, received);
                        OnReceive?.Invoke(new RemoteInfo
                        {
                            Socket = socket,
                            IsTcp = true,
                            IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6,
                            EndPoint = socket.RemoteEndPoint,
                            Data = data
                        });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }

            OnDisconnect?.Invoke(new RemoteInfo
            {
                Socket = socket,
                IsTcp = true,
                EndPoint = socket.RemoteEndPoint,
                IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6
            });
        }

        /// <summary>
        /// Send data to socket
        /// </summary>
        /// <param name="socket"></param>
        /// <param name="data"></param>
        /// <returns></returns>
        public async Task SendToAsync(Socket socket, byte[] data)
        {
            await socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None).ConfigureAwait(false);
            OnSend?.Invoke(new RemoteInfo
            {
                Socket = socket,
                IsTcp = true,
                EndPoint = socket.RemoteEndPoint,
                IsIpv6 = socket.AddressFamily == AddressFamily.InterNetworkV6,
                Data = data
            });
        }

        /// <summary>
        /// Stop TCP server
        /// </summary>
        /// <returns></returns>
        public async Task StopAsync()
        {
            _cancellationTokenSource.Cancel();
            _listener.Stop();
            if (_acceptTask != null)
            {
                try
                {
                    await _acceptTask.ConfigureAwait(false);
                }
                catch (AggregateException) { }
            }
        }
    }
}