﻿using System.Net.Sockets;
using System.Net;
using System.Threading.Tasks;
using System;
using System.Threading;

namespace EonaCat.Network
{
    public class SocketUdpClient
    {
        /// <summary>
        /// OnConnect event
        /// </summary>
        public event Action<EndPoint> OnConnect;

        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<byte[]> OnReceive;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<EndPoint> OnDisconnect;

        /// <summary>
        /// OnSend event
        /// </summary>
        public event Action<EndPoint, byte[]> OnSend;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        private UdpClient _udpClient;

        /// <summary>
        /// Create UDP client
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        /// <param name="cancellationToken"></param>
        public SocketUdpClient(IPAddress ipAddress, int port, CancellationToken cancellationToken = default)
        {
            CreateUdpClient(ipAddress, port, cancellationToken);
        }

        private void CreateUdpClient(IPAddress ipAddress, int port, CancellationToken cancellationToken = default)
        {
            IsIp6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            _udpClient = new UdpClient(ipAddress.AddressFamily);

            if (IsIp6)
            {
                _udpClient.Client.SetSocketOption(SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, false);
            }

            _udpClient.Client.Bind(new IPEndPoint(IsIp6 ? IPAddress.IPv6Any : IPAddress.Any, 0));

            if (IsMulticastGroupEnabled)
            {
                _udpClient.JoinMulticastGroup(ipAddress);
            }

            OnConnect?.Invoke(_udpClient.Client.RemoteEndPoint);
            _ = StartReceivingAsync(cancellationToken);
        }

        public bool IsMulticastGroupEnabled { get; set; }

        public bool IsIp6 { get; private set; }

        private async Task StartReceivingAsync(CancellationToken cancellationToken)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var result = await _udpClient.ReceiveAsync().ConfigureAwait(false);
                    OnReceive?.Invoke(result.Buffer);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }
            OnDisconnect?.Invoke(_udpClient.Client.RemoteEndPoint);
        }

        /// <summary>
        /// Send data to endPoint
        /// </summary>
        /// <param name="endPoint"></param>
        /// <param name="data"></param>
        /// <returns></returns>
        public async Task SendTo(EndPoint endPoint, byte[] data)
        {
            if (endPoint is IPEndPoint ipEndPoint)
            {
                await _udpClient.SendAsync(data, data.Length, ipEndPoint).ConfigureAwait(false);
                OnSend?.Invoke(endPoint, data);
            }
        }

        /// <summary>
        /// Disconnect UDP client
        /// </summary>
        public void Disconnect()
        {
            _udpClient.Close();
        }
    }
}
