﻿using System.Net.Sockets;
using System.Net;
using System;
using System.Runtime.InteropServices;
using System.Net.NetworkInformation;
using System.Threading.Tasks;
using System.Threading;

namespace EonaCat.Network
{
    public class SocketUdpServer
    {
        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<RemoteInfo> OnReceive;

        /// <summary>
        /// OnSend event
        /// </summary>
        public event Action<RemoteInfo> OnSend;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<RemoteInfo> OnDisconnect;

        private UdpClient udpClient;

        /// <summary>
        /// Determines if the UDP server is IpV6
        /// </summary>
        public bool IsIp6 { get; private set; }

        /// <summary>
        /// Create UDP server
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        public SocketUdpServer(IPAddress ipAddress, int port)
        {
            CreateUdpServer(ipAddress, port);
        }

        /// <summary>
        /// Create UDP server
        /// </summary>
        /// <param name="ipAddress"></param>
        /// <param name="port"></param>
        /// <exception cref="Exception"></exception>
        public SocketUdpServer(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            CreateUdpServer(ip, port);
        }

        private static void SetConnectionReset(Socket socket)
        {
            if (socket.ProtocolType != ProtocolType.Udp && !RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
            {
                return;
            }

            // Disable ICMP packet shutdown (forcibly closed)
            const int SioUdpConnReset = -1744830452;

            byte[] inValue = { 0, 0, 0, 0 };
            socket.IOControl(SioUdpConnReset, inValue, null);
        }

        private void CreateUdpServer(IPAddress ipAddress, int port)
        {
            IsIp6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            udpClient = new UdpClient(ipAddress.AddressFamily);
            SetConnectionReset(udpClient.Client);

            if (IsIp6)
            {
                udpClient.Client.SetSocketOption(SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, false);
            }

            udpClient.Client.Bind(new IPEndPoint(ipAddress, port));
        }

        /// <summary>
        /// Start UDP server
        /// </summary>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        public async Task StartAsync(CancellationToken cancellationToken = default)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    var result = await udpClient.ReceiveAsync().ConfigureAwait(false);
                    OnReceive?.Invoke(new RemoteInfo()
                    {
                        EndPoint = result.RemoteEndPoint,
                        IsIpv6 = result.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6,
                        Data = result.Buffer
                    });
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                }
            }
        }

        /// <summary>
        /// Send data to endPoint
        /// </summary>
        /// <param name="endPoint"></param>
        /// <param name="data"></param>
        /// <returns></returns>
        public async Task SendToAsync(EndPoint endPoint, byte[] data)
        {
            if (endPoint is IPEndPoint ipEndPoint)
            {
                await udpClient.SendAsync(data, data.Length, ipEndPoint).ConfigureAwait(false);
                OnSend?.Invoke(new RemoteInfo() { EndPoint = endPoint, Data = data, IsIpv6 = endPoint.AddressFamily == AddressFamily.InterNetworkV6 });
            }
        }

        /// <summary>
        /// Send data to all clients
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public async Task SendToAllAsync(byte[] data)
        {
            // get all connected clients
            var ipProperties = IPGlobalProperties.GetIPGlobalProperties();
            var endPoints = ipProperties.GetActiveUdpListeners();
            foreach (var endPoint in endPoints)
            {
                await udpClient.SendAsync(data, data.Length, endPoint).ConfigureAwait(false);
            }
        }
    }

}
