﻿using System;
using System.Net.WebSockets;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class WebSocketSecureClient
    {
        private const int BUFFER_SIZE = 4096;

        /// <summary>
        /// OnConnect event
        /// </summary>
        public event Action<RemoteInfo> OnConnect;

        /// <summary>
        /// OnReceive event
        /// </summary>
        public event Action<RemoteInfo> OnReceive;

        /// <summary>
        /// OnDisconnect event
        /// </summary>
        public event Action<RemoteInfo> OnDisconnect;

        /// <summary>
        /// OnError event
        /// </summary>
        public event Action<Exception, string> OnError;

        private ClientWebSocket _webSocket;

        /// <summary>
        /// The client name to be sent when connecting (optional).
        /// </summary>
        public string ClientName { get; set; }

        /// <summary>
        /// Create secure WebSocket client with certificate support
        /// </summary>
        /// <param name="uri"></param>
        /// <param name="clientCertificate">The client certificate for the connection</param>
        /// <param name="password">The password for the connection</param>
        /// <returns></returns>
        public Task ConnectAsync(string uri, X509Certificate2 clientCertificate = null, string password = null)
        {
            return CreateWebSocketClientAsync(uri, clientCertificate, password);
        }

        private async Task CreateWebSocketClientAsync(string uri, X509Certificate2 clientCertificate, string password)
        {
            _webSocket = new ClientWebSocket();

            if (clientCertificate != null)
            {
                _webSocket.Options.ClientCertificates.Add(clientCertificate);
            }

            if (!string.IsNullOrEmpty(password))
            {
                var passwordBytes = Encoding.UTF8.GetBytes(password);
                _webSocket.Options.SetRequestHeader("Password", Convert.ToBase64String(passwordBytes));
            }

            if (!string.IsNullOrEmpty(ClientName))
            {
                _webSocket.Options.SetRequestHeader("ClientName", ClientName);
            }

            try
            {
                Uri serverUri = new Uri(uri);
                await _webSocket.ConnectAsync(serverUri, CancellationToken.None).ConfigureAwait(false);
                OnConnect?.Invoke(new RemoteInfo { IsWebSocket = true });
                _ = StartReceivingAsync();
            }
            catch (Exception ex)
            {
                OnError?.Invoke(ex, $"Exception: {ex.Message}");
                Disconnect();
            }
        }

        private async Task StartReceivingAsync()
        {
            byte[] buffer = new byte[BUFFER_SIZE];
            while (_webSocket.State == WebSocketState.Open)
            {
                try
                {
                    var result = await _webSocket.ReceiveAsync(new ArraySegment<byte>(buffer), CancellationToken.None).ConfigureAwait(false);
                    if (result.Count > 0)
                    {
                        byte[] data = new byte[result.Count];
                        Buffer.BlockCopy(buffer, 0, data, 0, result.Count);
                        OnReceive?.Invoke(new RemoteInfo { IsWebSocket = true, Data = data });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (Exception ex)
                {
                    OnError?.Invoke(ex, $"Exception: {ex.Message}");
                    break;
                }
            }
            OnDisconnect?.Invoke(new RemoteInfo { IsWebSocket = true });
        }

        /// <summary>
        /// Send data
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public Task SendAsync(byte[] data)
        {
            return _webSocket.SendAsync(new ArraySegment<byte>(data), WebSocketMessageType.Binary, true, CancellationToken.None);
        }

        /// <summary>
        /// Disconnect
        /// </summary>
        public void Disconnect()
        {
            _webSocket?.CloseAsync(WebSocketCloseStatus.NormalClosure, "Disconnecting", CancellationToken.None);
        }
    }
}
